"""Cyclotomic rational functions.
"""

from sage.all import *

import itertools

from collections import Counter

from tmplist import TemporaryList

from util import monomial_exp, monomial_log, evaluate_polynomial, unzip, split_vector, common_overring

from util import create_logger
logger = create_logger(__name__)
        
class CyclotomicRationalFunction(object):
    """A cyclotomic rational function is an expression
    Q = f * X^alpha[0]/prod(1-X^alpha[i], i=1,..,k), where alpha[i] in ZZ^n
    and f is a polynomial in X = [X[0], ...].
    """

    __slots__ = [ 'polynomial', 'ring', 'variables', 'nvars', 'exponents',
                  '_known_to_be_normalised', '_known_to_be_reduced' ]

    def __init__(self, polynomial, exponents=None,
                 known_to_be_normalised=False,
                 known_to_be_reduced=False):

        # NOTE: polynomial.ring() must be 'multivariate' (even if there's just
        # one variable)

        self.polynomial = polynomial
        self.ring = self.polynomial.parent()
        self.variables = self.ring.gens()
        self.nvars = len(self.variables)

        if (exponents is None) or (polynomial == 0):
            self.exponents = [zero_vector(ZZ, self.nvars)]
            known_to_be_normalised = True
        else:
            self.exponents = [vector(ZZ,e) for e in exponents]

        self.exponents = [self.exponents[0]] + sorted(self.exponents[1:])

        if 0 in self.exponents[1:]:
            raise TypeError('Only non-zero vectors can be used as exponents for the denominator')

        if any(len(e) != self.nvars for e in self.exponents):
            raise TypeError('Exponent vector and number number of variables do not match')

        self._known_to_be_normalised = known_to_be_normalised
        self._known_to_be_reduced = known_to_be_reduced

    def copy(self):
        return CyclotomicRationalFunction(self.polynomial, self.exponents,
                                          self._known_to_be_normalised,
                                          self._known_to_be_reduced)

    def __nonzero__(self):
        return bool(self.polynomial)

    def degree(self, i=0):
        if not self.polynomial:
            return -Infinity
        x = self.variables[i]
        return self.polynomial.degree(x) + self.exponents[0][i] - sum(max(0, a[i]) for a in self.exponents[1:])

    def is_compatible_with(self, Q):
        # Check if we can add these two.
        return self.ring == Q.ring

    def evaluate(self, variables=None):
        if variables is None:
            R = self.ring
            K = FractionField(R)

            if not self.polynomial:
                return K.zero()

            variables = [K.coerce(x) for x in self.variables]
            exp = lambda v: K.prod(x**e for (x,e) in zip(variables, v))
            return K.product(
                evaluate_polynomial(self.polynomial, variables),
                exp(self.exponents[0]) / K.prod(K.one()-exp(e) for e in self.exponents[1:])
                )

        if variables == 'symbolic':
            variables = [SR(str(v)) for v in self.variables]
            coerc = lambda x: SR(x)
        else:
            coerc = lambda x: x

        exp = lambda v: prod(variables[i] ** v[i] for i in xrange(self.nvars))
        return coerc(
            evaluate_polynomial(self.polynomial, variables) *
            exp(self.exponents[0]) / prod(1-exp(e) for e in self.exponents[1:])
            )

    def evaluate_reciprocal(self):
        """Return 1/self.evaluate() assuming that it is a polynomial, that
        self.polynomial is constant, and that self.exponents[0] contains
        non-positive entries only.
        """

        if not self.polynomial.is_constant() or not self.polynomial:
            raise ValueError('polynomial is not a non-zero constant')

        R = self.ring
        exp = lambda v: R.prod(self.variables[i] ** v[i] for i in xrange(self.nvars))

        y, z = split_vector(self.exponents[0])
        if y:
            raise ValueError('not a polynomial')

        res = R.one()
        for c in self.exponents[1:]:
            a, b = split_vector(c)
            z -= b
            res = R.product(res, exp(b) - exp(a))

        if any(e < 0 for e in z):
            raise ValueError('not a polynomial')

        Q = R.base_ring()
        return R.prod([R.coerce(Q.one()/Q(self.polynomial)), exp(z), res])
   
    def triangulate(self):
        """Produce a generator yielding 'simplicial' cyclotomic rational
        functions (-> polynomials are constant) that sum up to 'self'.
        """
        coeff = self.polynomial.coefficients()
        mon = self.polynomial.monomials()

        for i in xrange(len(mon)):
            exps = self.exponents[:]
            exps[0] += monomial_log(mon[i])
            yield CyclotomicRationalFunction(self.ring(coeff[i]), exps).normalise()

    def __str__(self):
        return str(self.evaluate())

    def normalise(self):
        # 'self' is normalised iff
        # (1) the first non-zero entry in each ray in the denominator is positive and
        # (2) gcd(self.polynomial) == 1; over SR, ignore (2).
        
        if self._known_to_be_normalised or not self:
            return self

        f = self.polynomial
        alpha = self.exponents[:]
        k = len(alpha)-1

        for i in xrange(1, k+1): # we use f * X^u/(1-X^v) == (-f) * X^(u-v)/(1-X^(-v))
            if next(b for b in alpha[i] if b != 0) < 0:
                alpha[i] = -alpha[i]
                f = -f
                alpha[0] += alpha[i]

        try:
            g = gcd(f.monomials())
        except (AttributeError, NotImplementedError):
            # This doesn't currently work for polynomials over functions fields.
            pass
        else:
            f //= g
            alpha[0] += monomial_log(g)

        # NOTE: normalising a reduced rational function produces another
        # reduced one.
        return CyclotomicRationalFunction(
            f, exponents=alpha,
            known_to_be_normalised=True,
            known_to_be_reduced=self._known_to_be_reduced)

    def reduce(self):
        """Cancel factors in the denominator.
        """

        if self.ring.base_ring() == SR:
            return self # again, few things work in the symbolic case
        elif self._known_to_be_reduced:
            return self
        elif self.polynomial.is_constant():
            Q = self.copy()
            Q._known_to_be_reduced = True
            return Q

        f = self.polynomial
        alpha = self.exponents[:]
        k = len(alpha)-1

        exp = lambda v: self.ring.prod(self.variables[i] ** v[i] for i in xrange(self.nvars)) # only non-negative entries allowed here!
        
        for i in xrange(1, k + 1):
            beta = alpha[i]

            # (1) Check if we can cancel all of 1-X^beta.
            # We use f/(1-X^beta) = X^(beta^-)*f/(X^(beta^-)-X(^beta^+)
            # to translate between polynomials and Laurent polynomials.

            beta_pos, beta_neg = split_vector(beta)
            g = exp(beta_neg) - exp(beta_pos)
            if g.divides(f):
                f //= g
                alpha[0] += beta_neg
                alpha[i] = None
                continue

            # (2) Check if we can replace 1-X^beta by some proper divisor.
            e = gcd(beta)
            e = -e if e < 0 else e # NOTE: the documentation of Sage doesn't seem to guarantee non-negativity of GCDs
            if e == 1:
                continue

            gamma = beta / e

            for d in divisors(e):
                if d == e:
                    break

                delta = d * gamma
                m = e // d
                assert beta == m * delta

                delta_pos, delta_neg = split_vector(delta)
                y = exp(delta_pos)/exp(delta_neg) # == exp(delta)

                g = (self.ring)(exp((m-1) * delta_neg) * sum(y ** i for i in xrange(m)))

                # (1-X^beta)/(1-X^delta) = 1 + X^delta + ... + X^((m-1)*delta)
                #                        = X^(-(m-1)*delta^-) * g
                # Hence:
                # f/(1-X^beta) = (f/g) * X^((m-1)*delta^-) / (1-X^delta)

                if not g.divides(f):
                    continue

                f //= g
                alpha[i] = delta
                alpha[0] += (m-1) * delta_neg
                break

        return CyclotomicRationalFunction(
            f, exponents=[alpha[0]] + [a for a in alpha[1:] if a is not None],
            known_to_be_reduced=True)  # NOTE: The output need not be normalised.

    def simplify(self):
        return self.reduce().normalise()

    @staticmethod
    def common_denominator(it):
        """Construct a cyclotomic rational function 'cr' such that 'self'
        can be written as f * cr for a (non-Laurent) polynomial f.
        """

        a = None
        D = {}
        ring = None

        for cr in it:
            if ring is None:
                ring = cr.ring

            _, b = split_vector(cr.exponents[0])
            a = b if a is None else vector(ZZ, [max(x) for x in zip(a,b)])

            E = Counter([tuple(v) for v in cr.exponents[1:]])
            for v, m in E.items():
                D[v] = max(D[v],m) if v in D else m

        if a is None or ring is None:
            raise RuntimeError('need a non-empty iterable')

        exps = []
        for v, m in D.items():
            exps.extend([v] * m)
            _, vneg = split_vector(v)
            a += m * vneg
            
        return CyclotomicRationalFunction(ring.one(), [-a] + exps) # .normalise()

    def __neg__(self):
        Q = self.copy()
        Q.polynomial = -Q.polynomial
        return Q

    def __add__(self, other):
        if self.nvars != other.nvars:
            raise ValueError('inconsistent numbers of variables')

        ring = common_overring(self.ring, other.ring)

        if self.ring.variable_names() != other.ring.variable_names():
            raise ValueError('inconsistent variable names')

        if self.exponents[1:] == other.exponents[1:]:
            K = FractionField(ring)
            variables = [K.coerce(x) for x in self.variables]

            exp = lambda v: K.prod(variables[i] ** v[i] for i in xrange(self.nvars))
            return CyclotomicRationalFunction.from_laurent_polynomial(
                exp(self.exponents[0]) * K.coerce(self.polynomial) + exp(other.exponents[0]) * K.coerce(other.polynomial),
                ring,
                exponents=[vector(ZZ,self.ring.ngens())] + self.exponents[1:])
        elif not self:
            return other
        elif not other:
            return self

        Q = self.normalise()
        other = other.normalise()

        alpha = Q.exponents
        alpha_pos, alpha_neg = unzip(split_vector(v) for v in alpha)

        beta = other.exponents
        beta_pos, beta_neg = unzip(split_vector(w) for w in beta)
        
        k = len(alpha)-1
        ell = len(beta)-1

        I = range(1, k+1)
        J = range(1, ell+1)
        for i in xrange(1, k+1):
            j = next((j for j in J if alpha[i] == beta[j]), None)
            if j is not None:
                I.remove(i)
                J.remove(j)

        # Now I and J contain unmatched indices so the sum is given by
        #
        # X^alpha[0] * prod(1-X^beta[j], j in J) * f + X^beta[0] * prod(1-X^alpha[i], i in I) * g
        # ---------------------------------------------------------------------------------------
        # prod(1-X^alpha[i], i=1,..k) * prod(1-X^beta[j], j in J)
        #

        # Multiplying the numerator by exp(v+w) yields a polynomial.
        v = alpha_neg[0] + sum(beta_neg[j] for j in J)
        w = beta_neg[0] + sum(alpha_neg[i] for i in I)

        exp = lambda v: ring.prod(self.variables[i] ** v[i] for i in xrange(self.nvars))
        h = ring.summation(
            ring.prod( [ exp(w), exp(alpha_pos[0]) ] + [ exp(beta_neg[j])  -  exp(beta_pos[j]) for j in J ] + [ Q.polynomial ] ),
            ring.prod( [ exp(v),  exp(beta_pos[0]) ] + [ exp(alpha_neg[i]) - exp(alpha_pos[i]) for i in I ] + [ other.polynomial ] )
            )
        exponents = [-v-w] + alpha[1:] + [beta[j] for j in J]
        return CyclotomicRationalFunction(h, exponents).normalise()

    def __sub__(self, other):
        return self + (-other)

    def __mul__(self, other):
        if self.nvars != other.nvars:
            raise ValueError('inconsistent numbers of variables')
        if self.ring.variable_names() != other.ring.variable_names():
            raise ValueError('inconsistent variable names')
        ring = common_overring(self.ring, other.ring)

        return CyclotomicRationalFunction(
            ring.product(ring(self.polynomial), ring(other.polynomial)),
            [self.exponents[0] + other.exponents[0]] +
             self.exponents[1:] + other.exponents[1:]
            )

    def monomial_substitution(self, new_ring, Phi):
        """
        Perform a monomial substitution (assumed to be valid) of variables.

        'new_ring' should be a polynomial ring in, say, 'm' variables z_0, ...
        Let 'n' be the number of variables x_0, ... of 'self'.
        'Phi' should be an n by m matrix over ZZ, the rows of
        which encode the substitution x_i |--> z^Phi[i].
        """

        Phi = matrix(ZZ, Phi)
        m = new_ring.ngens()
        if (self.nvars != Phi.nrows()) or (m != Phi.ncols()):
            raise TypeError('The substitution matrix has the wrong size')

        # The rule x_i |--> z^Phi[i] yields x^a |--> z^(a * Phi).
        beta = [a * Phi for a in self.exponents]
        if 0 in beta[1:]:
            raise ZeroDivisionError('Invalid monomial substitution')

        K = FractionField(new_ring)
        new_variables = [K.coerce(x) for x in new_ring.gens()]
        exp = lambda a: K.prod(x ** e for (x,e) in zip(new_variables,a))
        h = evaluate_polynomial(self.polynomial, [exp(v) for v in Phi])
        return CyclotomicRationalFunction.from_laurent_polynomial(h, new_ring, exponents=beta)
        
    @classmethod
    def from_laurent_polynomial(cls, f, ring, exponents=None):
        # NOTE:
        # We currently use rational functions instead of either Sage's
        # built-in Laurent polynomials or those from laurent.py.

        K = FractionField(ring)
        variables = [K.coerce(x) for x in ring.gens()]

        if exponents is None:
            exponents = [vector(ZZ, len(variables))]

        try:
            f = K(f)
        except TypeError:
            raise TypeError('Not a rational function over the given ring')

        den = f.denominator()
        lc = den.lc()
        num = f.numerator() / lc
        den = den / lc
        assert num == ring(num) 
        assert (den == 1) or den.is_monomial()
        try:
            return cls(num, [-monomial_log(den) + exponents[0]] + exponents[1:])
        except TypeError:
            raise TypeError('Not a Laurent polynomial')

    @classmethod
    def from_split_expression(cls, g, y, ring):
        # Suppose 'g' is a symbolic expression in the variables 'y' such
        # that the denominator of 'g' splits into linear factors 1-y[i].
        # Turn 'g' into a cyclotomic rational function over 'ring'.
        # Here, y[i] will be turned into ring.gens()[i]

        num = g.numerator()
        r = len(y)
        e = identity_matrix(QQ, r)

        exponents = [zero_vector(len(y))]

        # Unpack the denominator with multiplicities.
        for pol in itertools.chain.from_iterable([a] * int(i) for (a, i) in g.denominator().factor_list()):
            # Look for 'pol + 1' in y. If it isn't there, then 'pol' should be a constant.
            i = next((i for i in xrange(r) if y[i] == pol + 1), None)
            if i is None:
                # Move constants into the numerator
                if not ((PolynomialRing(QQ,y))(pol)).is_constant():
                    raise TypeError('Cannot handle this expression')
                num /= pol
            else:
                num = -num # Encode (y[i]-1) as (-1) * (1-y[i]).
                exponents.append(e[i])

        return cls.from_laurent_polynomial(num({ a: ring.gen(i) for i,a in enumerate(y)}), ring, exponents=exponents)




"""This module provides basic functionality for toric data.
"""

from sage.all import *

from convex import conify_polyhedron, \
    is_contained_in_dual, dual_cone_as_polyhedron, \
    inner_open_normal_fan, get_point_in_polyhedron, \
    DirectProductOfPolyhedra, linear_image_of_polyhedron

from util import normalise_poly, monomial_log, \
    subpolynomial_meet, is_subpolynomial, initial_form_by_direction, \
    unzip, cached_simple_method, minimal_elements, split_off_torus

import common
import itertools

from util import create_logger
logger = create_logger(__name__)

def relative_initial_forms(F, polyhedron=None, split=True):
    if not F:
        raise ValueError('need non-empty collection of polynomials')

    if split:
        G, d, A = split_off_torus(F)
        N = sum(g.newton_polytope() for g in G)
        n = A.nrows()
    else:
        N = sum(f.newton_polytope() for f in F)
        n = N.ambient_dim()

    if polyhedron is None:
        polyhedron = Polyhedron(ieqs=[(n+1)*(0,)])

    for Q in inner_open_normal_fan(N):
        if split:
            Q = linear_image_of_polyhedron(DirectProductOfPolyhedra(Q, Polyhedron(ieqs=[(n-d+1)*(0,)])),
                                           A.transpose())
        Q &= polyhedron
        if Q.dim() == -1:
            continue
        y = get_point_in_polyhedron(Q)
        yield [initial_form_by_direction(f, y) for f in F], Q

    
def belongs_to_radical(f, I):
    """Test if f in I.radical().
    """

    R = PolynomialRing(QQ, I.ring().ngens()+1, I.ring().variable_names() + ('Zoo',))
    Zoo = R.gens()[-1]
    J = R.ideal([R(g) for g in I.gens()] + [R(1) - Zoo*R(f)])
    return [R(1)] == J.groebner_basis(algorithm='singular' if common.plumber else '')

def is_nondegenerate(F, all_subsets=True, all_initial_forms=True, collision_handler=None):
    # NOTE: Setting all_subsets=False gives exactly Khovanskii's original
    # non-degeneracy condition.

    logger.debug('Original polynomials: %s' % F)
    if not F:
        return True

    if 0 in F:
        raise NotImplementedError()

    for G, _ in relative_initial_forms(F) if all_initial_forms else [(F,None)]:
        # NOTE: we could do the torus factor splitting for each subset separately.
        G, _, _ = split_off_torus(G)
        logger.debug('Restricted polynomials: %s' % G)

        R = G[0].parent()
        jac = jacobian(G, R.gens())

        idx = [i for i in xrange(len(G)) if not G[i].is_monomial()]
        if not all_subsets and len(idx) < len(G):
            continue # empty variety!

        logger.debug('Active indices: %s' % idx)
        S = Subsets(idx) if all_subsets else [idx]

        for J in S:
            if not J:
                continue
            I = R.ideal([G[j] for j in J] + matrix(R,[jac[j] for j in J]).minors(len(J)))
            if not belongs_to_radical(prod(R.gens()), I):
                logger.debug('Collision: %s' % J)
                if collision_handler is not None:
                    collision_handler(J)
                return False
    return True

class ToricDatum(object):
    """A toric datum consists of
    (1) a rational polyhedron in some valuation space,
    (2) an integrand given by two Laurent vectors encoding Laurent monomials,
    (3) polynomial divisibility conditions with monomial left-hand sides
        Laurent polynomials representing, and
    (4) partial markings of initial forms among the right-hand sides.
    """
    def __init__(self, ring, integrand, cc, initials, polyhedron, depth=0):
        """Initialise a toric datum.

        'ring' should be a polynomial ring over QQ, the number of variables of which
        is taken to be the ambient dimension.

        When evaluating associated integrals, we will sum over
        t^(integrand[0] * y) * p^(integrand[1] * y) with y in polyhedron.

        'cc' should be a list-like object consisting of pairs (m,f) with 'm'
        monomial and 'f' in 'ring'.

        Everything is copied and normalised. We may further replace the given
        toric datum by an equivalent one.
        
        If it is set, 'initials' should be a list-like object s.t. initials[i]
        is either 'None' or a sub-polynomial of cc[i][1]. It is then assumed
        but not checked that initials[i] is the initial form of cc[i][1]
        throughout 'polyhedron'.
        """

        # Bookkeeping

        self.ring = ring
        self.ambient_dim = ring.ngens()
        self.integrand = tuple(integrand)
        self.polyhedron = polyhedron
        self._coll_idx = None
        self._ordered = False
        self._depth = depth

        # Empty polyhedra can easily arise as part of reduction
        if self.polyhedron.dim() == -1:
            self.cc, self.lhs, self.rhs, self.initials = [], [], [], []
            return

        if cc is None:
            cc = []
        initials = len(cc)*[None] if initials is None else initials[:]

        if len(cc) != len(initials):
            raise TypeError('The number of initials has to be the same as the number of conditions')

        # Discard conditions 'm || 0'.
        # We shouldn't remove 'constant || f' unless we know that everything
        # is integral.
        # This code fails if '0' is a LHS which should not be allowed anyway.
        idx = [i for i in xrange(len(cc)) if (cc[i][1] != Integer(0))] # and (not cc[i][0].is_constant())]
        cc, initials = [cc[i] for i in idx], [initials[i] for i in idx]

        # Normalise everything. If initials are given, we use their leading terms
        # instead of those of the full RHS.
        # We also set the initials of monomials.
        self.lhs, self.rhs, self.initials = [], [], []
        for i in xrange(len(cc)):
            self.lhs.append(normalise_poly( cc[i][0]) )
            if initials[i] is None:
                f = normalise_poly( cc[i][1] )
                self.rhs.append(f)
                self.initials.append( f if f.is_monomial() else None )
            else:
                self.initials.append( initials[i]/initials[i].lc() )
                self.rhs.append( cc[i][1]/initials[i].lc() )
            
        self.cc = zip(self.lhs, self.rhs)
        
        # Some sanity checks.
        for i in xrange(len(self.cc)):
            if not self.lhs[i].is_monomial():
                raise TypeError('LHS has to consist of monomials')
            if not(self.initials[i] is None) and not(is_subpolynomial(self.rhs[i], self.initials[i])):
                raise TypeError('initial forms need to be sub-polynomials')
            
    def length(self):
        """Return the number of divisibility conditions.
        """
        return len(self.cc)

    def weight(self):
        return sum(len(r.monomials())-1 for r in self.rhs)

    def _purge_multiples(self):
        if self.is_empty():
            return self

        D = conify_polyhedron(self.polyhedron)
        F = FractionField(self.ring)
        def lte(i,j):
            if i == j:
                return True
            rat = F(self.rhs[j]*self.lhs[i]/self.rhs[i]/self.lhs[j])
            try:
                alpha = monomial_log(rat.numerator()) - monomial_log(rat.denominator())
            except ValueError:
                return False
            return is_contained_in_dual(Cone([alpha]), D)

        mins = minimal_elements(range(len(self.rhs)), lte)
        if len(mins) == len(self.rhs):
            return self
        return ToricDatum(self.ring, self.integrand,
                                  [(self.lhs[i],self.rhs[i]) for i in mins],
                                  [self.initials[i] for i in mins],
                                  self.polyhedron, depth=self._depth)


    def simplify(self):
        """Simplify a toric datum.
        
        We do the following:
        (1) Move pseudo-monomial divisibility conditions into the polyhedron.
        (2) Perform support reduction using the polyhedron.
        (3) Scan for divisibility relations.
        """

        try:
            if self._is_simple:
                return self
        except AttributeError:
            pass

        # T is the current toric datum
        T = self

        cnt = 0
        while true:
            cnt += 1
            if cnt > 2:
                logger.info('Simplification entered round %d' % cnt) # e.g. triggered for Perm(123)

            logger.debug('Looping. Current datum:\n%s' % T)

            # T = T.divide_out_variables()

            changed = false
            polyhedron = T.polyhedron
            new_cc, new_initials = [], []

            # Step 1: Move pseudo-monomial conditions into the polyhedron.
            # We say that a condition is pseudo-monomial if it's initial form is monomial.

            idx_mon = [i for i in xrange(len(T.cc)) if (T.initials[i] is not None) and T.initials[i].is_monomial()]
            idx_nonmon = [i for i in xrange(len(T.cc)) if not i in idx_mon]
            logger.debug('Number of conditions: %d (pseudo-monomials: %d)' % (len(T.cc),len(idx_mon)))

            if len(idx_mon) > 0:
                logger.debug('Shrinking polyhedron.')

                polyhedron &= dual_cone_as_polyhedron([monomial_log(T.initials[i]) - monomial_log(T.lhs[i]) for i in idx_mon])

                # NOTE: we don't set 'changed = True' at this point. No need.

            # Step 2: 
            # For each non-monomial condition, remove terms divisible by the
            # LHS within the polyhedron
            cone = conify_polyhedron(polyhedron)
            for i in idx_nonmon:  # or 'xrange(len(T.cc))', to be safe
                lhs, rhs = T.cc[i] ; init = T.initials[i]

                g = gcd(lhs,rhs)
                lhs, rhs = lhs//g, rhs//g
                if not(init is None):
                    init //= g

                # Find redundant terms of 'rhs'.
                mon, coeff = rhs.monomials(), rhs.coefficients()
                idx_red = [ j for j in xrange(len(mon))
                            if is_contained_in_dual(Cone([monomial_log(mon[j]) - monomial_log(lhs)], lattice=ZZ**T.ambient_dim), cone) ]

                # Case a: No term of 'rhs' is redundant
                if len(idx_red) == 0:
                    new_cc.append((lhs, rhs)) ; new_initials.append(init)
                    continue

                changed = true

                # Case b: (lhs || rhs) follows from the monomial conditions.
                if len(idx_red) == len(mon):
                    logger.debug('Redundant condition: %s || %s' % (lhs,rhs))
                    continue
                
                # Case c: lhs || rhs can be simplified but not removed.
                logger.debug('Condition %s || %s becomes...')

                red = sum(coeff[k]*mon[k] for k in idx_red) # = redundant form of 'rhs'
                rhs -= red

                g = gcd(lhs,rhs)
                lhs, rhs = lhs//g, rhs//g

                if not(init is None):
                    # NOTE:
                    # If init != None and some term of init is redundant, then the
                    # entire condition 'lhs || rhs' has to be redundant.
                    if subpolynomial_meet(init, red) != Integer(0):
                        raise RuntimeError('Impossible redundancy. This should never happen.')

                    init = (init - subpolynomial_meet(init, red))//g
                    init = None if init==Integer(0) else init

                    # Sanity check.
                    if not(is_subpolynomial(rhs, init)):
                        raise RuntimeError('I messed up an initial form')
                    
                logger.debug('... %s || %s' % (lhs,rhs))
                new_cc.append((lhs,rhs)) ; new_initials.append(init)

            T = ToricDatum(ring=T.ring, integrand=T.integrand,
                                    cc=new_cc, initials=new_initials,
                                    polyhedron=polyhedron, depth=self._depth)

            if not changed:
                break

        T = T._purge_multiples()
        T._is_simple = True
        return T

    def is_empty(self):
        return self.polyhedron.dim() == -1

    @cached_simple_method
    def is_balanced(self):
        """Check if the toric datum is known to be balanced.
        """
        return None not in self.initials

    def is_ordered(self):
        return self._ordered

    def balance(self, strategy='min'):
        if self.is_balanced():
            yield self
            return

        idx, weights = unzip((i,len(f.monomials())) for i,f in enumerate(self.rhs) if self.initials[i] is None)

        if strategy == 'full':
            J = idx
        elif strategy == 'min':
            J = [ idx[weights.index(min(weights))] ]
        elif strategy == 'max':
            J = [ idx[weights.index(max(weights))] ]
        else:
            raise TypeError('unknown balancing strategy')

        for F, Q in relative_initial_forms([self.rhs[j] for j in J], self.polyhedron):
            new_initials = self.initials[:]
            for (j,f) in itertools.izip(J, F):
                new_initials[j] = f
            yield ToricDatum( ring=self.ring, integrand=self.integrand,
                              cc=self.cc, initials=new_initials,
                              polyhedron=Q, depth=self._depth )

    def order(self):
        if self._ordered:
            yield T
            return

        if not self.is_balanced():
            raise ValueError('Cannot order unbalanced conditions')

        alpha = [ monomial_log(self.initials[i].monomials()[0])-monomial_log(self.lhs[i])
                  for i in range(len(self.rhs)) ]

        I = range(len(self.rhs)) if not self._coll_idx else self._coll_idx
        for pi in itertools.permutations(I, len(I)):
            polyhedron = self.polyhedron
            for i in xrange(len(I)-1):
                polyhedron &= dual_cone_as_polyhedron([alpha[pi[i+1]]-alpha[pi[i]]],
                                                      strict=pi[i] > pi[i+1])

            T = ToricDatum( ring=self.ring, integrand=self.integrand,
                                    cc=self.cc, initials=self.initials,
                                    polyhedron=polyhedron, depth=self._depth )
            T = T.simplify()

            if T.is_empty():
                continue
            T._ordered = True
            yield T

    @cached_simple_method
    def is_regular(self):
        """Test if a balanced toric datum is regular.
        """

        if self.is_empty():
            return True
        if not self.is_balanced():
            return False
        return is_nondegenerate(self.initials, all_subsets=True, all_initial_forms=False,
                                collision_handler=lambda J: setattr(self, '_coll_idx', J))

    def reduce(self, i, j, term_i=None, term_j=None, strict=False):
        """
        For terms term_i and term_j of RHS[i] and RHS[j], respectively, enforce
        term_i || term_j and simplify conditions.
        By default, leading terms are used. If strict is 'True', then strict
        divisibility is enforced.
        We do not require the toric datum to be balanced.
        """

        if i == j:
            raise RuntimeError('I can only reduce pairs of distinct conditions')

        if term_i is None:
            term_i = self.rhs[i].lt() if self.initials[i] is None else self.initials[i].lt()
        else:
            if not is_subpolynomial(self.rhs[i], term_i):
                raise ValueError('Not a term of rhs[i]')
        if term_j is None:
            term_j = self.rhs[j].lt() if self.initials[j] is None else self.initials[j].lt()
        else:
            if not is_subpolynomial(self.rhs[i], term_i):
                raise ValueError('Not a term of rhs[j]')

        lhs_i = self.lhs[i]
        lhs_j = self.lhs[j]

        logger.debug('Reduction: (%d) --> (%d)' % (i,j))
        aa = lhs_j * term_i ; bb = lhs_i * term_j
        gg = gcd(aa,bb); aa //= gg ; bb //= gg
        aa //= aa.lc() ; bb //= bb.lc()

        if strict:
            logger.debug('Enforcing %s || %s (strict divisibility)' % (aa, bb))
        else:
            logger.debug('Enforcing %s || %s ' % (aa,bb))

        polyhedron = self.polyhedron & dual_cone_as_polyhedron([monomial_log(bb) - monomial_log(aa)], strict=strict)

        if polyhedron.dim() == -1:
            logger.debug('Found empty reduction.')
            return ToricDatum( ring=self.ring, integrand=self.integrand,
                                       cc=None, initials=None,
                                       polyhedron=polyhedron, depth=self._depth )


        lhs, rhs, initials = self.lhs[:], self.rhs[:], self.initials[:]

        R = self.ring ;  K = R.fraction_field()

        g = K( self.rhs[j]/self.lhs[j] - term_j/(self.lhs[j] * term_i) * self.rhs[i] )
        lhs[j], rhs[j] = g.denominator(), g.numerator()

        logger.debug('Condition (%d) becomes %s || %s.' % (j, str(lhs[j]), str(rhs[j])))

        if (initials[i] is None) or (initials[j] is None):
            initials[j] = None
        else:
            # Unless the above transformation yields zero on initial forms, we can
            # read off the new one in position j. 
            h = K( self.initials[j]/self.lhs[j] - term_j/(self.lhs[j] * term_i) * self.initials[i] )
            initials[j] = None if h == K(0) else R( h * g.denominator() )

            if not(initials[j] is None) and not is_subpolynomial(rhs[j], initials[j]):
                # If term_i and term_j are INITIAL terms, this can't happen!
                raise RuntimeError('This reduction is not admissible')

            logger.debug('New intial form of (%d) is %s.' % (j, 'undetermined' if initials[j] is None else str(initials[j])))

        return ToricDatum( ring=self.ring, integrand=self.integrand,
                                   cc=zip(lhs,rhs), initials=initials,
                                   polyhedron=polyhedron, depth=self._depth )

    def __str__(self):
        S = 'Toric datum\n'\
            'Base ring: ' + str(self.ring) + '\n\n'\
            'Polyhedron: ' + str(self.polyhedron) + '\n' +\
            str(self.polyhedron.cdd_Hrepresentation()) + '\n'\
            'Integrand: t^(y * ' + str(self.integrand[0]) + ') * q^(y * ' + str(self.integrand[1]) + ')\n'

        if len(self.cc) == 0:
            return S + 'No divisibility conditions.'

        width = max(map(lambda f: len(filter( lambda i: i > 0, f.exponents()[0]) ),self.lhs)) * 4 + 2
        for i in xrange(len(self.cc)):
            if self.initials[i] is None:
                S += ('(%2d )  %'+str(width)+'s  ||  %s\n') % (i, str(self.lhs[i]), str(self.rhs[i]))
            else:
                S += ('(%2d )  %'+str(width)+'s  || <<%s>>') % (i, str(self.lhs[i]), str(self.initials[i]))
                rem = self.rhs[i] - self.initials[i]
                if rem.lc()  > Integer(0):
                    S += ' + ' + str(rem)
                elif rem.lc() < Integer(0):
                    S += ' ' + str(rem)
                S += '\n'
        return S



from sage.all import *

from convex import RationalSet, dual_cone_as_polyhedron, PositiveOrthant
import util
import toric

import itertools

class LaurentError(Exception):
    pass

from util import create_logger
logger = create_logger(__name__)

class LaurentPolynomial(object):
    __slots__ = ['ring', 'nvars', 'den', 'num', 'f']

    def __init__(self, f):
        # NOTE: f should belong to a *multivariate* polynomial ring over
        # a field or to its field of fractions.
        f = FractionField(f.parent())(f)
        self.ring = f.parent()
        self.nvars = self.ring.ngens()
        self.den = f.denominator()
        self.num = f.numerator()/self.den.lc()
        self.den //= self.den.lc()
        if not self.den.is_monomial():
            raise LaurentError('not a Laurent polynomial')
        self.f = self.num/self.den

    def decompose(self):
        beta = vector(self.den.exponents()[0])
        return [(c,vector(QQ,alpha)-beta) for c,alpha in zip(self.num.coefficients(), self.num.exponents())]

    def exponents(self):
        return [alpha for _,alpha in self.decompose()]

    def terms(self):
        return [LaurentPolynomial(c*t/self.den) for (c,t) in list(self.num)]

    def is_integer_valued(self, RS=None):
        if RS is None:
            # Test if self.f is an honest polynomial.
            return self.den == 1
        return all(RS.is_contained_in_dual(alpha) for alpha in self.exponents())

    def is_term(self):
        return len(self.exponents()) == 1

    def is_unit(self, RS=None):
        li = self.exponents()
        if len(li) != 1:
            return False
        if RS is None:
            return True
        return RS.is_perpendicular(li[0])

    def divides(self, other):
        try:
            _ = other/self
        except LaurentError:
            return False
        else:
            return True

    def is_subpolynomial(self, other):
        li = other.decompose()
        return all(a in li for a in self.decompose())

    def factor(self):
        return Factorization([(LaurentPolynomial(g),e) for (g,e) in self.f.factor()])

    def __nonzero__(self):
        return bool(self.f)

    def __add__(self, other):
        return LaurentPolynomial(self.f + other.f)

    def __sub__(self, other):
        return LaurentPolynomial(self.f - other.f)

    def __mul__(self, other):
        return LaurentPolynomial(self.f * other.f)

    def __div__(self, other):
        return LaurentPolynomial(self.f / other.f)

    def __pow__(self, exponent):
        return LaurentPolynomial(self.f ** exponent)

    def __eq__(self, other):
        return self.f == other.f

    def __hash__(self):
        return hash(self.f)

    def __str__(self):
        return str(self.f)

    def __repr__(self):
        return repr(self.f)

class LaurentIdeal(object):
    # Ideals of the rings of RS-integer Laurent polynomials.
    # Detects units, removes zeros and multiples.

    __slots__ = ['ring', 'gens', 'initials', 'RS']

    def __init__(self, gens, initials=None, RS=None, ring=None, normalise=False):
        initials = [None for f in gens] if initials is None else list(initials)

        if not gens and ring is None:
            raise ValueError('need to specify base ring')
        self.ring = ring if ring is not None else gens[0].ring

        if RS is None:
            RS = RationalSet(PositiveOrthant(self.ring.ngens()))

        if RS.ambient_dim != self.ring.ngens():
            raise ValueError('rational set and number of variables are inconsistent')

        if RS.is_empty():
            # Q: is this controversial?
            self.gens = []
            self.initials = []
            self.RS = RS
            return

        # Get rid of zeros, detect units.
        new_gens, new_initials = [], []
        for f, init in zip(gens, initials):
            if not f:
                continue
            if f.is_unit(RS):
                new_gens = [LaurentPolynomial(f.ring(1))]
                new_initials = [LaurentPolynomial(f.ring(1))]
                break
            new_gens.append(f)

            if f.is_term():
                foo = f
            elif init is None:
                foo = None
            else:
                foo = init
            new_initials.append(foo)

        # Now get rid of g whenever g/f is an RS-integer-valued
        # Laurent polynomial.
        def lte(i,j):
            if i == j:
                return True
            try:
                g = new_gens[j]/new_gens[i]
            except LaurentError:
                return False
            return g.is_integer_valued(RS)
        mins = util.minimal_elements(range(len(new_gens)), lte)

        self.gens = [new_gens[i] for i in mins]
        self.initials = [new_initials[i] for i in mins]
        self.RS = RS

        if normalise:
            for i, f in enumerate(self.gens):
                c = LaurentPolynomial(f.ring(f.num.lc()))
                self.gens[i] /= c
                if self.initials[i] is None:
                    continue
                self.initials[i] /= c

        # BEGIN SANITY CHECK
        assert all(
            (self.initials[i] is None) or self.initials[i].is_subpolynomial(self.gens[i])
            for i in xrange(len(self.gens)))
        # END SANITY CHECK

    def __eq__(self, other):
        raise NotImplementedError

    def simplify(self):
        I = LaurentIdeal(gens=self.gens, initials=self.initials, RS=self.RS, ring=self.ring, normalise=True)

        changed = True
        while changed:
            changed = False
            gens = I.gens[:]
            initials = I.initials[:]
            for i,j in itertools.product(range(len(gens)), repeat=2):
                if i == j:
                    continue
                for ti,tj in itertools.product(
                        (gens[i] if initials[i] is None else initials[i]).terms(),
                        (gens[j] if initials[j] is None else initials[j]).terms()):
                    quo = ti/tj
                    if not quo.is_integer_valued(self.RS):
                        continue
                    f = gens[i] - quo*gens[j]
                    if len(f.exponents()) < len(gens[i].exponents()):
                        changed = True
                        gens[i] = f
                        if (initials[i] is not None) and (initials[j] is not None):
                            initials[i] = initials[i] - quo * initials[j]
                            if not initials[i]:
                                initials[i] = None
                        else:
                            initials[i] = None
                        I = LaurentIdeal(gens=gens, initials=initials, RS=I.RS, ring=I.ring, normalise=True)
                        break
                else:
                    continue
                break
        return I

    def reduce(self, i, j, ti, tj, strict):
        # Enforce t_i || t_j and change F.gens[j] in a new Laurent ideal.
        # For balanced polynomials, initial terms should be used.

        if i == j:
            raise ValueError('need distinct indices')
        if not (ti in self.gens[i].terms() and tj in self.gens[j].terms()):
            raise ValueError('not a term of the given specified polynomial')
        if ((self.initials[i] is not None) and (ti not in self.initials[i].terms())) or ((self.initials[j] is not None) and (tj not in self.initials[j].terms())):
            raise ValueError('non-initial term')

        alpha_i = ti.exponents()[0]
        alpha_j = tj.exponents()[0]
        RS = self.RS & RationalSet(dual_cone_as_polyhedron([alpha_j - alpha_i], strict=strict))

        gens = self.gens[:]
        initials = self.initials[:]

        gens[j] = gens[j] - tj/ti * gens[i]
        if (initials[i] is None) or (initials[j] is None):
            initials[j] = None
        else:
            f = initials[j] - tj/ti * initials[i]
            initials[j] = f if f else None
        return LaurentIdeal(gens=gens, initials=initials, RS=RS, ring=self.ring)
        
    def is_balanced(self):
        return not any(init is None for init in self.initials)
    
    def balance(self, strategy='min'):
        # NOTE 1: repeated application might be necessary 
        # NOTE 2: shares code with ToricDatum.balance
        if self.is_balanced():
            yield self
            return
        idx, weights = zip(*[(i, len(self.gens[i].exponents())) for i in range(len(self.gens)) if self.initials[i] is None])
        J = [idx[weights.index(min(weights))]] if strategy == 'min' else idx

        for F, Q in toric.relative_initial_forms([self.gens[j].num for j in J], polyhedron=None):
            new_RS = self.RS & RationalSet(Q)
            if new_RS.is_empty():
                continue
            new_initials = self.initials[:]
            for j,f in zip(J, F):
                new_initials[j] = LaurentPolynomial(f/self.gens[j].den)
            yield LaurentIdeal(gens=self.gens, initials=new_initials, RS=new_RS, ring=self.ring)

    def __str__(self):
        return 'LaurentIdeal(gens=%s, initials=%s)' % (self.gens, self.initials)
    __repr__ = __str__



"""
Operations for sums of rational functions and an interface to LattE's 'count --multivariate-generating-function'
"""

from sage.all import *
from sage.features.latte import Latte_count
from sage.modules.free_module_element import vector

import common

from util import my_find_executable, TemporaryDirectory, cd, augmented_env
from cycrat import CyclotomicRationalFunction

from tmplist import TemporaryList

import itertools

import subprocess

from util import create_logger
logger = create_logger(__name__)

if not Latte_count().is_present():
    logger.warn('LattE/count not found. '
                'Computations of p-adic zeta functions are unavailable. '
                'You can try to install it by running "sage -i latte_int"')

def NonnegativeCompositions(n, length=None):
    # Produce all k-lists of nonnegative integers that sum up to k
    for c in Compositions(n+length, length=length):
        yield [a-1 for a in c]

def get_totally_nonperp_vector(vectors, strategy='random'):
    """Construct a vector 'w' such that w * v != 0 for all v in vectors.
    """
    vectors = list(vectors) # We want to allow generators.

    if not vectors:
        return None

    n = len(vectors[0])
    for k in (k for k in itertools.count() for _ in xrange((k+1) * n)):
        if strategy == 'random':
            v = random_vector(n, x=-k, y=k+2)
        elif strategy == 'moment':
            v = vector(ZZ, [1] + [k ** i for i in xrange(1,n)])
        else:
            raise TypeError('unknown strategy')

        if 0 in (v * w for w in vectors):
            continue
        return v

def taylor_processor_naive(new_ring, Phi, scalar, alpha, I, omega):
    k = alpha.nrows() - 1
    tau = SR.var('tau')
    y = [SR('y%d' % i) for i in xrange(k+1)]

    R = PolynomialRing(QQ, len(y), y)
    beta = [a * Phi for a in alpha]

    J = [i for i in xrange(1, k+1) if not(i in I)]

    def f(i):
        if i == 0:
            return QQ(scalar) * y[0] * exp(tau * omega[0])
        elif i in I:
            return 1/(1 - exp(tau * omega[i]))
        else:
            return 1/(1 - y[i] * exp(tau * omega[i]))

    h = prod(f(i) for i in xrange(k+1))

    # Get constant term of h as a Laurent series in tau.
    g = h.series(tau, 1).truncate().collect(tau).coefficient(tau, 0)
    g = g.factor() if g else g
    yield CyclotomicRationalFunction.from_split_expression(g, y, R).monomial_substitution(new_ring, beta)
    
def taylor_processor_factored(new_ring, Phi, scalar, alpha, I, omega):
    k = alpha.nrows() - 1
    tau = SR.var('tau')
    y = [SR('y%d' % i) for i in xrange(k+1)]

    R = PolynomialRing(QQ, len(y), y)
    beta = [a * Phi for a in alpha]
    
    ell = len(I)
    def f(i):
        if i == 0:
            return QQ(scalar) * y[0] * exp(tau * omega[0])
        elif i in I:
            return tau/(1 - exp(tau * omega[i]))
        else:
            return 1/(1 - y[i] * exp(tau * omega[i]))

    H = [f(i).series(tau, ell+1).truncate().collect(tau) for i in xrange(k+1)]

    for i in xrange(k+1):
        H[i] = [H[i].coefficient(tau, j) for j in xrange(ell+1)]

    r = []

    # Get coefficient of tau^ell in prod(H)

    for w in NonnegativeCompositions(ell, k+1):
        r = prod(
            CyclotomicRationalFunction.from_split_expression(H[i][w[i]], y, R).monomial_substitution(new_ring, beta)
            for i in xrange(k+1))
        yield r

def latteify_polyhedron(P):
    res = []
    lin = []

    for s in P.cdd_Hrepresentation().splitlines():
        s = s.strip()
        if s in ['H-representation', 'begin', 'end']:
            continue
        elif s.find('rational') != -1:
            res.append(s[:-len(' rational')])
        elif s.find('linearity') != -1:
            lin.append(s)
        else:
            res.append(s)
    return '\n'.join(res + lin) + '\n'

class SMURF(object):
    """
    Sums of MUltivariate Rational Functions.
    """

    def __init__(self, arg, base_list=None):
        # A sufficiently list-like object 'base_list' can be provided;
        # otherwise, we just use a native list.

        self.summands = [] if base_list is None else base_list
        if isinstance(arg, sage.rings.ring.CommutativeRing):
            self.ring = arg
        elif isinstance(arg, CyclotomicRationalFunction):
            self.summands.append(arg.copy())
            self.ring = arg.ring
        else: # we're expecting a non-empty iterable of CyclotomicRationalFunctions
            self.summands.extend(a.copy() for a in arg)
            self.ring = self.summands[0].ring

        if not self.__is_consistent():
            raise TypeError('These rational functions do not belong together')

    def __is_consistent(self):
        return all(self.summands[0].is_compatible_with(a)
                   for a in self.summands)

    def __iter__(self):
        return iter(self.summands)

    def __add__(self, other):
        if other == 0:
            other = SMURF(self.ring) # this allows us to use 'sum' for SMURFs

        if (not self.summands) and (not other.summands):
            if self.ring != other.ring:
                raise TypeError('Different rings')
            return SMURF(self.ring)
        return SMURF(itertools.chain(self.summands, other.summands))

    __radd__ = __add__

    def extend(self, other):
        if self.ring != other.ring:
            raise TypeError('Different rings')

        self.summands.extend(other.summands)
        if not self.__is_consistent():
            raise TypeError

    __iadd__ = extend
    
    def append(self, cr):
        self.summands.append(cr)
        if not self.__is_consistent():
            raise TypeError
        
    def __str__(self):
        return 'Sum of %d cyclotomic rational functions over %s' % (len(self.summands), self.ring.gens())

    def evaluate(self, variables=None):
        # This is included for debugging purposes!
        return sum(SR(s.evaluate(variables)) for s in self.summands)

    def monomial_substitution(self, new_ring, Phi, base_list=None, taylor_processor=None):
        """
        Perform monomial substitutions which are valid for the sum of 'self'
        but perhaps not for each summand.
        The algorithm used can be found in Lemma 2.5 and Theorem 2.6 of
        Barvinok, Woods: ``Short rational generating functions for lattice point
        problems'', JAMS (2003).
        """

        # NOTE: 
        # we only ever apply this function to sums which compute an integral.

        if taylor_processor is None:
            taylor_processor = taylor_processor_factored # taylor_processor_naive

        if not(self.summands):
            return SMURF(new_ring, base_list=base_list)

        v = get_totally_nonperp_vector(
            vector(QQ, w) for w in itertools.chain.from_iterable(
                f.exponents[1:] for f in self.summands
                )
            )

        Phi = matrix(QQ, Phi)
        L = Phi.column_space()
        Lperp = L.basis_matrix().right_kernel() 

        with TemporaryList() as res:
            for f in itertools.chain.from_iterable(s.triangulate() for s in self.summands):
                # First, try to apply the substitution directly. Only if that fails,
                # do we use the far more involved method of Barvinok & Woods.
                try:
                    res.append(f.monomial_substitution(new_ring, Phi))
                    continue
                except ZeroDivisionError:
                    pass

                # Setting: f == scalar * X^alpha[0] / prod(1 - X^alpha[i], i=1..k)
                scalar, alpha, k = f.polynomial, matrix(QQ,f.exponents), len(f.exponents)-1 # Note the final '-1'!
                if not scalar:
                    continue
                assert scalar.is_constant()  # note the use of 'triangulate' above

                omega = [v*a for a in alpha]
                I = [i for i in xrange(1, k+1) if alpha[i] in Lperp]
                res.extend(taylor_processor(new_ring, Phi, scalar, alpha, I, omega))

            return SMURF(new_ring, base_list=base_list) if not res else SMURF(res, base_list=base_list)

    @classmethod
    def from_polyhedron(cls, P, ring, base_list=None):
        """
        Use LattE to compute the generating function of a rational polyhedron
        as a sum of small rational functions.
        """
        from sage.misc.parser import Parser

        if P.is_empty():
            if ring.ngens() != P.ambient_dim():
                raise TypeError('Dimension mismatch')
            return cls(ring, base_list=base_list)
        elif P.is_zero():
            return cls([CyclotomicRationalFunction(ring.one())], base_list=base_list)
        elif len(P.vertices()) == 1 and (not P.rays()) and (not P.lines()):
            # For some reason, LattE doesn't produce .rat files for points.
            return cls([CyclotomicRationalFunction(ring.one(), exponents=[vector(ZZ,P.vertices()[0])])])

        hrep = 'polyhedron.hrep';
        ratfun = hrep + '.rat';

        with TemporaryDirectory() as tmpdir, cd(tmpdir):
            with open(hrep, 'w') as f:
                f.write(latteify_polyhedron(P))

            with open('/dev/null', 'w') as DEVNULL:
                retcode = subprocess.call([common.count,
                                           '--compute-vertex-cones=4ti2',
                                           '--triangulation=cddlib',
                                           '--multivariate-generating-function',
                                           hrep],
                                          stdout=DEVNULL, stderr=DEVNULL,
                                          env=augmented_env(common.count))
            if retcode != 0:
                raise RuntimeError('LattE failed. Make sure it has been patched in order to be compatible with Zeta.')

            K = FractionField(ring)
            Kone = K.one()
            pick_gen = {"x{}".format(i) : gen for (i, gen) in enumerate(K.gens())}
            maple_parser = Parser(make_var=pick_gen)

            def get_exponent(cycterm):
                monomial = -maple_parser.parse(cycterm) + Kone
                return (vector(monomial.numerator().exponents()[0]) -
                        vector(monomial.denominator().exponents()[0]))

            with TemporaryList() as summands, open(ratfun, 'r') as f:
                for line in f:
                    line = line.translate(None, "[]").strip()
                    if not line:
                        continue # ignore empty lines

                    numerator_line, denominator_line = line.split("/")
                    numerator = maple_parser.parse(numerator_line)

                    # denominator_line is of the form
                    #((1-x[0]*x[5])*(1-x[4])*(1-x[0]^(-1)*x[3])*(1-x[2]))
                    # So we remove the "outer" paranthesis, and split on ")*(".
                    # TODO: Find less hacky way to do this parsing.
                    denominator_line = denominator_line[1:-1]
                    if denominator_line.startswith("("):
                        denominator_line = denominator_line[1:]
                    if denominator_line.endswith(")"):
                        denominator_line = denominator_line[:-1]
                    den_list = denominator_line.split(")*(")
                    exponents = ([vector(ZZ, K.ngens())] +
                                [get_exponent(cycterm) for cycterm in den_list])

                    summands.append(CyclotomicRationalFunction.from_laurent_polynomial(
                                    numerator, ring, exponents))

                return cls(summands, base_list=base_list)



## -*- encoding: utf-8 -*-
import sys
import subprocess

from setuptools import setup
from codecs import open # To open the README file with proper encoding
from setuptools.command.test import test as TestCommand # for tests
from setuptools.extension import Extension
from Cython.Distutils import build_ext

# Get information from separate files (README, VERSION)
def readfile(filename):
    with open(filename, encoding='utf-8') as f:
        return f.read()

# For the tests
class SageTest(TestCommand):
    def run_tests(self):
        subprocess.check_call(["sage", "-t", "--force-lib", "docs/source", "zetalib"])

# Hack to make sure `make build` is called before bdist_wheel.
try:
    from wheel.bdist_wheel import bdist_wheel
    class MakeBuild(bdist_wheel):
        def run(self):
            subprocess.check_call(["make", "build"])
            bdist_wheel.run(self)
except ImportError:
    # TODO: Change to logging.warn(...) and let MakeBuild = None
    sys.exit("The wheel package is needed for bdist_wheel to work. "
             "It can be installed by running `sage -pip install wheel`")

setup(
    name = "zetalib",
    version = readfile("VERSION").strip(), # the VERSION file is shared with the documentation
    description='computing zeta functions of groups, algebras, and modules',
    long_description = readfile("README.rst"), # get the long description from the README
    url='https://gitlab.com/mathzeta2/zetalib',
    author='Tobias Rossmann, Tomer Bauer',
    license='GPLv3+', # This should be consistent with the LICENSE file
    classifiers=[
      # How mature is this project? Common values are
      #   3 - Alpha
      #   4 - Beta
      #   5 - Production/Stable
      'Development Status :: 4 - Beta',
      'Intended Audience :: Science/Research',
      'Topic :: Software Development :: Build Tools',
      'Topic :: Scientific/Engineering :: Mathematics',
      'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
      'Programming Language :: Python :: 2.7',
    ], # classifiers list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
    keywords = "SageMath packaging",
    packages = ['zetalib'],
    package_data = {'zetalib': ['addmany.pyx',
                                'crunch.*',
                                'p-adic-examples.json',
                                'Zeta.html',
                                ]},
    cmdclass = {
        'test': SageTest, # adding a special setup command for tests
        'bdist_wheel': MakeBuild,
        'build_ext': build_ext, # For cython files
    },
    ext_modules = [Extension("zetalib.addmany", ["zetalib/addmany.pyx"], include_dirs=['zetalib/'])],
)



"""
Basic functionality for 'additively free ZZ-algebras', i.e. non-associative
ring structures on ZZ^n.
"""

from sage.all import *
from convex import PositiveOrthant
from toric import ToricDatum
from util import normalise_poly, monomial_log, cached_simple_method, \
    upper_triangular_matrix, is_block_diagonal_matrix, basis_of_matrix_algebra

import common
import multiprocessing

import itertools

#TODO: Use SageObject instead of object.
class Algebra(object):
    """Additively free non-associative ZZ-algebras with optional operators.
    """

    def __init__(self, table=None, rank=None, blocks=None, operators=None,
                 matrix=None, bilinear=False, simple_basis=False,
                 descr=None, matrix_basis=None, product=None):

        """Construct an additively free ZZ-algebra (possibly with operators).
        """

        self.descr = descr
        self.operators = [] if (operators is None) else [Matrix(QQ,x) for x in operators]

        if not(matrix is None):
            A = Matrix(QQ,matrix)
            if not(bilinear or simple_basis):
                raise TypeError('You need to set either bilinear or simple_basis')

            if simple_basis:
                self.rank = A.ncols()
                e = MatrixSpace(QQ, self.rank).identity_matrix().rows()
                def f(i,j):
                    if abs(A[i,j]) > self.rank:
                        raise TypeError('Matrix entries out of range')
                    if   A[i,j] == 0: return (Integer(0) * e[0])
                    elif A[i,j] > 0:  return (e[A[i,j]-1])
                    elif A[i,j] < 0:  return (-e[-A[i,j]-1])
                self.table = [ [ f(i,j) for j in xrange(self.rank) ] for i in xrange(self.rank) ]
            elif bilinear:
                n = A.ncols()
                self.rank = n + Integer(1)
                w = vector(QQ, n * [Integer(0)] + [Integer(1)])
                self.table = [ [ A[i,j] * w for j in xrange(n) ] + [ 0 * w ] for i in xrange(n) ] + [ [ 0 * w for j in xrange(n+1) ] ]

            self.blocks = (self.rank,) if blocks is None else tuple(blocks)
            return
                
        # If the rank is set, always initialise an empty table.
        if not(rank is None):
            self.rank = rank
            self.table = [ [ vector(QQ, self.rank) for j in xrange(self.rank) ] for i in xrange(self.rank) ]
            self.blocks = (self.rank,) if blocks is None else tuple(blocks)

        if isinstance(table, dict):
            if rank is None:
                raise TypeError('You need to specify the rank') 
            for pos, v in table.iteritems():
                self.table[pos[0]][pos[1]] = vector(QQ, v)
            return

        if isinstance(table, list):
            self.rank = len(table)

            self.blocks = (self.rank,) if blocks is None else tuple(blocks)
            self.table = [ [ vector(QQ, v) for v in row ] for row in table ]
            return
        
        # 'basis' is a list of QQ-linearly independent matrixy things
        # spanning a Lie or associative algebra.
        if not(matrix_basis is None):
            if product is None:
                raise TypeError('need to specify Lie or associative product')
            if product == 'Lie':
                mult = lambda x, y: x*y - y*x
            elif product == 'associative':
                mult = lambda x, y: x*y
            else:
                raise TypeError('unknown type of product')

            self.rank = len(matrix_basis)
            self.blocks = (self.rank,) if blocks is None else tuple(blocks)

            matrix_basis = [Matrix(QQ, b) for b in matrix_basis]
            A = Matrix(QQ, [b.list() for b in matrix_basis])

            if A.rank() != len(matrix_basis):
                raise TypeError('matrices specified are linearly dependent')

            self.table = [ [ A.solve_left(vector(mult(b,c).list())) for c in matrix_basis] for b in matrix_basis ]
            return

        if rank is None:
            raise ValueError('need to specify the rank')

    def multiply(self, v, w, ring=QQ):
        return vector(ring, 
                      [ sum(v[i] * w[j] * ring(self.table[i][j][k]) for i in xrange(self.rank) for j in xrange(self.rank)) for k in xrange(self.rank) ]
                      )

    def right_multiplication(self, w, ring=QQ):
        return matrix(ring, [self.multiply(v,w) for v in identity_matrix(ring, self.rank)])
                      
    def ideal(self, vecs, ring=QQ):
        V = ring**self.rank
        B = list(identity_matrix(QQ,self.rank))
    
        U = V.subspace(vecs)
        while True:
            li = list(U.basis())
            xi = []
            for u,e in itertools.product(li,B):
                xi.append(self.multiply(u,e,ring=ring))
                xi.append(self.multiply(e,u,ring=ring))
            W = V.subspace(li+xi)
            if U == W:
                return U
            else:
                U = W

    @cached_simple_method
    def derived_series(self, ring=QQ):
        V = ring**self.rank
        dseries = [V]
        while True:
            W = self.ideal([self.multiply(v,w,ring=ring) for v in V.basis() for w in V.basis()])
            if V == W:
                return dseries
            else:
                V = W
                dseries.append(V)

    @cached_simple_method
    def derived_length(self):
        dseries = self.derived_series()
        return Infinity if dseries[-1].dimension() > 0 else len(dseries)-1

    @cached_simple_method
    def is_soluble(self):
        return self.derived_length() < Infinity

    @cached_simple_method
    def lower_central_series(self, ring=QQ):
        V = ring**self.rank
        B = list(V.basis())
        lcseries = [V]
        while True:
            li = []
            for v,e in itertools.product(V.basis(), B):
                li.append(self.multiply(v,e))
                li.append(self.multiply(e,v))
            W = self.ideal(li)
            if V == W:
                return lcseries
            else:
                V = W
                lcseries.append(V)

    @cached_simple_method
    def nilpotency_class(self):
        lcseries = self.lower_central_series()
        return Infinity if lcseries[-1].dimension() > 0 else len(lcseries)-1

    @cached_simple_method
    def is_nilpotent(self):
        return self.nilpotency_class() < Infinity

    @cached_simple_method
    def is_commutative(self):
        for u,v in itertools.product((QQ**self.rank).basis_matrix(),repeat=2):
            if self.multiply(u,v) != self.multiply(v,u):
                return False
        return True

    @cached_simple_method
    def is_anticommutative(self):
        for u,v in itertools.product((QQ**self.rank).basis_matrix(),repeat=2):
            if self.multiply(u,v) != -self.multiply(v,u):
                return False
        return True

    @cached_simple_method
    def is_associative(self):
        for u,v,w in itertools.product((QQ**self.rank).basis_matrix(),repeat=3):
            if self.multiply(self.multiply(u,v),w) != self.multiply(u,self.multiply(v,w)):
                return False
        return True

    @cached_simple_method
    def is_Lie(self):
        if not self.is_anticommutative():
            return False
        for u,v,w in itertools.product((QQ**self.rank).basis_matrix(),repeat=3):
            if self.multiply(self.multiply(u,v),w) != self.multiply(self.multiply(u,w),v) + self.multiply(u,self.multiply(v,w)):
                return False
        return True

    # Use rows of a matrix as the new basis.
    def change_basis(self, A, check=True):
        if check and not is_block_diagonal_matrix(A, self.blocks):
            raise ValueError('the given matrix does not preserve the block structure')

        A = matrix(QQ, A)
        if not A.is_invertible():
            raise TypeError('The argument must be an invertible matrix over QQ.') 
        return Algebra( table = [[ self.multiply( A[i], A[j] ) * A.inverse() for j in xrange(self.rank) ] for i in xrange(self.rank)],
                        operators = [ A * op * A**(-1) for op in self.operators ], blocks = self.blocks )

    @cached_simple_method
    def _Lie_centre(self):
        # Construct the QQ-space of all z with a*z == z*a == 0 for all a.
        Z = QQ**self.rank
        for i in xrange(self.rank):
            Z = Z.intersection(
                matrix(QQ, [self.table[i][j] for j in xrange(self.rank)]).left_kernel()       
            )
            if self.is_anticommutative() or self.is_commutative():
                continue
            Z = Z.intersection(
                matrix(QQ, [self.table[j][i] for j in xrange(self.rank)]).left_kernel()
            )
        return Z

    @cached_simple_method
    def _adjusted_basis(self):
        # Construct a basis of 'self' as in (Stasinski & Voll 2014, Section 2.2.2);
        # we ignore finitely many primes.
        if len(self.blocks) > 1:
            raise NotImplementedError

        if not (self.is_Lie() and self.is_nilpotent() and self.rank > 0):
            raise NotImplementedError()

        Z = self._Lie_centre()
        D = self.derived_series()[1]

        M = Z.intersection(D)
        A = Z.intersection(M.complement())
        B = D.intersection(M.complement())
        C = (B+M+A).complement()
        li = list(itertools.chain(C.basis(), B.basis(), M.basis(), A.basis()))
        return matrix(QQ, [v.denominator()*v for v in li]), (dim(C), dim(B), dim(M), dim(A))

    @cached_simple_method
    def _adjoint_representation(self):
        # Construct a basis of the image of `self' under its adjoint
        # representation.
        assert self.is_anticommutative()
        return basis_of_matrix_algebra([self.right_multiplication(e) for e in identity_matrix(QQ, self.rank)],
                                       product='zero')
        
    @cached_simple_method
    def _commutator_matrices(self):
        if len(self.blocks) > 1:
            raise NotImplementedError

        A, dims = self._adjusted_basis()
        d = dims[1] + dims[2] # = dim of commutator ideal
        r = dims[0] + dims[1] # = codim of centre
        ring = PolynomialRing(QQ, 'y', d)
        y = vector(ring, ring.gens())

        T = self.change_basis(A)

        a = dims[0] ; b = a + d # [a,b) == indices of basis vectors of the commutator ideal

        R = matrix(ring, [
            [ y * vector(ring, [T.table[i][j][k] for k in xrange(a,b)])
              for j in xrange(r) ]
            for i in xrange(r) ]
                  )
        # get last dims[1] columns
        S = (R.transpose()[r-dims[1]:r]).transpose()
        return R, S

    def toric_datum(self, objects='subalgebras', name='x'):
        """Construct a toric datum associated with the enumeration of subobjects.

        We can enumerate subalgebras as well as left/right/2-sided ideals.
        """

        if not isinstance(objects, str):
            raise TypeError('String expected')
        if not objects in ['subalgebras', 'ideals', 'left ideals', 'right ideals']:
            raise TypeError('Unknown objects.')

        if sum(self.blocks) != self.rank:
            raise ValueError('blocks and rank do not match')

        nvars = sum(binomial(a+1,2) for a in self.blocks)
        R = PolynomialRing(QQ, nvars, name)
        entries = iter(R.gens())

        M = block_diagonal_matrix([upper_triangular_matrix(R, a, entries) for a in self.blocks])
        D = prod(M.diagonal())

        diag = iter(M.diagonal())
        zoo = prod(next(diag)**(i+1) for a in self.blocks for i in xrange(a))
        integrand = (monomial_log(D), monomial_log(zoo) - vector(QQ,R.ngens() * [1]))

        Madj = M.adjoint()
        E = identity_matrix(R, self.rank)
        li = []
        def insert_components(w):
            for rhs in w:
                if rhs.is_zero():
                    continue
                cand = normalise_poly(rhs)
                if not cand in li:
                    li.append(cand)

        for i in xrange(self.rank):
            for j in xrange(self.rank):
                if objects == 'subalgebras':
                    insert_components(self.multiply(M[i], M[j], R) * Madj)
                if (objects == 'right ideals') or (objects == 'ideals'):
                    insert_components(self.multiply(M[i], E[j], R) * Madj)
                if (objects == 'left ideals') or (objects == 'ideals'):
                    insert_components(self.multiply(E[j], M[i], R) * Madj)

        for op in self.operators:
            for row in M * matrix(R, op) * Madj:
                insert_components(row)

        if len(li) > 0:
            li = ideal(li).interreduced_basis()
        cc_raw = [(D,f) for f in li]

        # Perform some trivial simplifications
        cc = []
        for c in cc_raw:
            lhs, rhs = c
            if rhs.is_zero():
                continue
            g = gcd(lhs, rhs)
            rhs = rhs//g
            lhs = lhs//g
            if lhs.is_constant():
                continue
            mon, coeff = rhs.monomials(), rhs.coefficients()
            # Discard monomials in 'rhs' that are divisible by 'lhs'
            idx = [k for k in xrange(len(mon)) if not lhs.divides(mon[k])]
            rhs = sum(coeff[k] * mon[k] for k in idx)

            if rhs.is_zero():
                continue
            cand = (normalise_poly(lhs), normalise_poly(rhs))
            if not cand in cc:
                cc.append(cand) 
        cc.sort()
        T = ToricDatum(ring=R, integrand=integrand, cc=cc,
                               initials=None,
                               polyhedron=PositiveOrthant(R.ngens())
                               )
        return T.simplify() if T.weight() < common._simplify_bound else T

    def find_good_basis(self, objects='subalgebras', name='x'):
        """Find a basis yielding a 'good' toric datum.
        
        Presently we only look at permutations of the given basis
        (preserving the block structure, if any).
        """

        def weight(g):
            A = g.matrix()
            L = self.change_basis(A)
            return L.toric_datum(objects=objects, name=name).weight()

        Sn = SymmetricGroup(self.rank)
        partial_sums = [sum(self.blocks[:i]) for i in xrange(len(self.blocks)+1)]
        gens = []
        for r in xrange(len(partial_sums)-1):
            a = partial_sums[r]
            b = partial_sums[r+1]
            G = SymmetricGroup(range(a+1,b+1)) # NOTE: SymmetricGroup(n) acts on [1,...,n] instead of range(n).
            f = G.hom(Sn)
            gens.extend(f(g) for g in G.gens())
        G = Sn.subgroup(gens)

        best_weight, best_permutation = Infinity, None
        for (arg,w) in (parallel(ncpus=common.ncpus)(weight))(list(G)):
            g = arg[0][0]
            if w == 'NO DATA':
                w = weight(g)
            if w < best_weight:
                best_weight, best_permutation = w, g
        return matrix(QQ, best_permutation.matrix())
    
    def __str__(self):
        if self.descr:
            S = self.descr + '\n'
        else:
            S = "Additively free ZZ-algebra of rank " + str(self.rank) + " with multiplication table\n"

        if len(self.blocks) > 1:
            S = S + 'Blocks: %s\n' % list(self.blocks)

        for row in self.table:
            S = S + '\t'
            for vec in row:
                S = S + str(vec) + ', '
            S = S[:-2] + '\n'

        if len(self.operators) > 0:
            S += 'Operators: ' + str(len(self.operators)) + '\n'
            
        return S[:-1]

    def __repr__(self):
        return '%s(**%s)' % (self.__class__, self.__dict__)

def tensor_with_duals(L):
    if len(L.blocks) > 1:
        raise NotImplementedError

    n = L.rank
    shift = n * [n]
    zero  = n * [0]

    def f(i,j):
        if i < n and j < n:
            return list(L.table[i][j]) + zero
        elif i < n and j >= n:
            return zero + list(L.table[i][j-n])
        elif i >= n and j < n:
            return zero + list(L.table[i-n][j])
        else:
            return zero + zero
    return Algebra(table=[[f(i,j) for j in xrange(2*n)] for i in xrange(2*n)])

def tensor_with_3duals(L):
    if len(L.blocks) > 1:
        raise NotImplementedError

    n = L.rank
    shift = n * [n]
    zero  = n * [0]

    def f(i,j):
        if i < n and j < n:
            return list(L.table[i][j]) + zero + zero 
        elif i < n and n <= j < 2*n:
            return zero + list(L.table[i][j-n]) + zero
        elif i < n and 2*n <= j:
            return zero + zero + list(L.table[i][j-2*n])
        elif n <= i < 2*n and j < n:
            return zero + list(L.table[i-n][j]) + zero
        elif n <= i < 2*n and n <= j < 2*n:
            return zero + zero + list(L.table[i-n][j-n])
        elif 2*n <= i and j < n:
            return zero + zero + list(L.table[i-2*n][j])
        else:
            return zero + zero + zero
    return Algebra(table=[[f(i,j) for j in xrange(3*n)] for i in xrange(3*n)])
